@@ -1,6 +1,6 @@
 # Licensed under the GPL: https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
-# For details: https://github.com/PyCQA/pylint/blob/main/LICENSE
-# Copyright (c) https://github.com/PyCQA/pylint/blob/main/CONTRIBUTORS.txt
+# For details: https://github.com/pylint-dev/pylint/blob/main/LICENSE
+# Copyright (c) https://github.com/pylint-dev/pylint/blob/main/CONTRIBUTORS.txt
 
 """Puts the check_parallel system under test."""
 
@@ -9,24 +9,36 @@
 from __future__ import annotations
 
 import argparse
-import multiprocessing
 import os
+import sys
+from concurrent.futures import ProcessPoolExecutor
+from concurrent.futures.process import BrokenProcessPool
+from pathlib import Path
+from pickle import PickleError
+from typing import TYPE_CHECKING
+from unittest.mock import patch
 
 import dill
 import pytest
-from astroid import nodes
 
 import pylint.interfaces
 import pylint.lint.parallel
 from pylint.checkers import BaseRawFileChecker
-from pylint.lint import PyLinter
+from pylint.checkers.imports import ImportsChecker
+from pylint.lint import PyLinter, augmented_sys_path
 from pylint.lint.parallel import _worker_check_single_file as worker_check_single_file
 from pylint.lint.parallel import _worker_initialize as worker_initialize
 from pylint.lint.parallel import check_parallel
 from pylint.testutils import GenericTestReporter as Reporter
+from pylint.testutils.utils import _test_cwd
 from pylint.typing import FileItem
 from pylint.utils import LinterStats, ModuleStats
 
+if TYPE_CHECKING:
+    from unittest.mock import MagicMock
+
+    from astroid import nodes
+
 
 def _gen_file_data(idx: int = 0) -> FileItem:
     """Generates a file to use as a stream."""
@@ -109,7 +121,7 @@ class ParallelTestChecker(BaseRawFileChecker):
         for _ in self.data[1::2]:  # Work on pairs of files, see class docstring.
             self.add_message("R9999", args=("From process_module, two files seen.",))
 
-    def get_map_data(self):
+    def get_map_data(self) -> list[str]:
         return self.data
 
     def reduce_map_data(self, linter: PyLinter, data: list[list[str]]) -> None:
@@ -160,10 +172,10 @@ class ThirdParallelTestChecker(ParallelTestChecker):
 class TestCheckParallelFramework:
     """Tests the check_parallel() function's framework."""
 
-    def setup_class(self):
+    def setup_class(self) -> None:
         self._prev_global_linter = pylint.lint.parallel._worker_linter
 
-    def teardown_class(self):
+    def teardown_class(self) -> None:
         pylint.lint.parallel._worker_linter = self._prev_global_linter
 
     def test_worker_initialize(self) -> None:
@@ -171,25 +183,44 @@ class TestCheckParallelFramework:
         worker_initialize(linter=dill.dumps(linter))
         assert isinstance(pylint.lint.parallel._worker_linter, type(linter))
 
+    def test_worker_initialize_with_package_paths(self) -> None:
+        linter = PyLinter(reporter=Reporter())
+        with augmented_sys_path([]):
+            worker_initialize(
+                linter=dill.dumps(linter), extra_packages_paths=["fake-path"]
+            )
+            assert "fake-path" in sys.path
+
+    def test_worker_initialize_reregisters_custom_plugins(self) -> None:
+        linter = PyLinter(reporter=Reporter())
+        linter.load_plugin_modules(["pylint.extensions.private_import"])
+
+        pickled = dill.dumps(linter)
+        with patch(
+            "pylint.extensions.private_import.register", side_effect=AssertionError
+        ):
+            with pytest.raises(AssertionError):
+                worker_initialize(linter=pickled)
+
     @pytest.mark.needs_two_cores
     def test_worker_initialize_pickling(self) -> None:
         """Test that we can pickle objects that standard pickling in multiprocessing can't.
 
         See:
         https://stackoverflow.com/questions/8804830/python-multiprocessing-picklingerror-cant-pickle-type-function
-        https://github.com/PyCQA/pylint/pull/5584
+        https://github.com/pylint-dev/pylint/pull/5584
         """
         linter = PyLinter(reporter=Reporter())
         linter.attribute = argparse.ArgumentParser()  # type: ignore[attr-defined]
-        with multiprocessing.Pool(
-            2, initializer=worker_initialize, initargs=[dill.dumps(linter)]
-        ) as pool:
-            pool.imap_unordered(print, [1, 2])
+        with ProcessPoolExecutor(
+            max_workers=2, initializer=worker_initialize, initargs=(dill.dumps(linter),)
+        ) as executor:
+            executor.map(print, [1, 2])
 
     def test_worker_check_single_file_uninitialised(self) -> None:
         pylint.lint.parallel._worker_linter = None
         with pytest.raises(  # Objects that do not match the linter interface will fail
-            Exception, match="Worker linter not yet initialised"
+            RuntimeError, match="Worker linter not yet initialised"
         ):
             worker_check_single_file(_gen_file_data())
 
@@ -231,6 +262,29 @@ class TestCheckParallelFramework:
         assert stats.statement == 18
         assert stats.warning == 0
 
+    def test_linter_with_unpickleable_plugins_is_pickleable(self) -> None:
+        """The linter needs to be pickle-able in order to be passed between workers."""
+        linter = PyLinter(reporter=Reporter())
+        # We load an extension that we know is not pickle-safe
+        linter.load_plugin_modules(["pylint.extensions.overlapping_exceptions"])
+        try:
+            dill.dumps(linter)
+            # TODO: 4.0: Fix this test by raising this assertion again
+            # raise AssertionError(
+            #     "Plugins loaded were pickle-safe! This test needs altering"
+            # )
+        except (KeyError, TypeError, PickleError, NotImplementedError):
+            pass
+
+        # And expect this call to make it pickle-able
+        linter.load_plugin_configuration()
+        try:
+            dill.dumps(linter)
+        except KeyError as exc:
+            raise AssertionError(
+                "Cannot pickle linter when using non-pickleable plugin"
+            ) from exc
+
     def test_worker_check_sequential_checker(self) -> None:
         """Same as test_worker_check_single_file_no_checkers with SequentialTestChecker."""
         linter = PyLinter(reporter=Reporter())
@@ -300,7 +354,6 @@ class TestCheckParallel:
             linter,
             jobs=1,
             files=iter(single_file_container),
-            arguments=["--enable", "R9999"],
         )
         assert len(linter.get_checkers()) == 2, (
             "We should only have the 'main' and 'sequential-checker' "
@@ -329,6 +382,7 @@ class TestCheckParallel:
         # now run the regular mode of checking files and check that, in this proc, we
         # collect the right data
         filepath = [single_file_container[0][1]]  # get the filepath element
+        linter.stats = LinterStats()
         linter.check(filepath)
         assert {
             "input.similar1": {  # module is the only change from previous
@@ -366,7 +420,9 @@ class TestCheckParallel:
         # Invoke the lint process in a multi-process way, although we only specify one
         # job.
         check_parallel(
-            linter, jobs=1, files=iter(single_file_container), arguments=None
+            linter,
+            jobs=1,
+            files=iter(single_file_container),
         )
 
         assert {
@@ -411,7 +467,9 @@ class TestCheckParallel:
             (10, 2, 3),
         ],
     )
-    def test_compare_workers_to_single_proc(self, num_files, num_jobs, num_checkers):
+    def test_compare_workers_to_single_proc(
+        self, num_files: int, num_jobs: int, num_checkers: int
+    ) -> None:
         """Compares the 3 key parameters for check_parallel() produces the same results.
 
         The intent here is to ensure that the check_parallel() operates on each file,
@@ -421,7 +479,6 @@ class TestCheckParallel:
         This test becomes more important if we want to change how we parameterize the
         checkers, for example if we aim to batch the files across jobs.
         """
-
         # define the stats we expect to get back from the runs, these should only vary
         # with the number of files.
         expected_stats = LinterStats(
@@ -478,7 +535,6 @@ class TestCheckParallel:
                     linter,
                     jobs=num_jobs,
                     files=file_infos,
-                    arguments=None,
                 )
                 stats_check_parallel = linter.stats
                 assert linter.msg_status == 0, "We should not fail the lint"
@@ -508,14 +564,13 @@ class TestCheckParallel:
             (10, 2, 3),
         ],
     )
-    def test_map_reduce(self, num_files, num_jobs, num_checkers):
+    def test_map_reduce(self, num_files: int, num_jobs: int, num_checkers: int) -> None:
         """Compares the 3 key parameters for check_parallel() produces the same results.
 
         The intent here is to validate the reduce step: no stats should be lost.
 
-        Checks regression of https://github.com/PyCQA/pylint/issues/4118
+        Checks regression of https://github.com/pylint-dev/pylint/issues/4118
         """
-
         # define the stats we expect to get back from the runs, these should only vary
         # with the number of files.
         file_infos = _gen_file_datas(num_files)
@@ -546,9 +601,117 @@ class TestCheckParallel:
                     linter,
                     jobs=num_jobs,
                     files=file_infos,
-                    arguments=None,
                 )
                 stats_check_parallel = linter.stats
+        # pylint: disable=possibly-used-before-assignment
         assert str(stats_single_proc.by_msg) == str(
             stats_check_parallel.by_msg
         ), "Single-proc and check_parallel() should return the same thing"
+
+    @pytest.mark.timeout(5)
+    def test_no_deadlock_due_to_initializer_error(self) -> None:
+        """Tests that an error in the initializer for the parallel jobs doesn't
+        lead to a deadlock.
+        """
+        linter = PyLinter(reporter=Reporter())
+
+        linter.register_checker(SequentialTestChecker(linter))
+
+        # Create a dummy file, the actual contents of which will be ignored by the
+        # register test checkers, but it will trigger at least a single-job to be run.
+        single_file_container = _gen_file_datas(count=1)
+
+        # The error in the initializer should trigger a BrokenProcessPool exception
+        with pytest.raises(BrokenProcessPool):
+            check_parallel(
+                linter,
+                jobs=1,
+                files=iter(single_file_container),
+                # This will trigger an exception in the initializer for the parallel jobs
+                # because arguments has to be an Iterable.
+                extra_packages_paths=1,  # type: ignore[arg-type]
+            )
+
+    @pytest.mark.needs_two_cores
+    def test_cyclic_import_parallel(self) -> None:
+        tests_dir = Path("tests")
+        package_path = Path("input") / "func_w0401_package"
+        linter = PyLinter(reporter=Reporter())
+        linter.register_checker(ImportsChecker(linter))
+
+        with _test_cwd(tests_dir):
+            check_parallel(
+                linter,
+                jobs=2,
+                files=[
+                    FileItem(
+                        name="input.func_w0401_package.all_the_things",
+                        filepath=str(package_path / "all_the_things.py"),
+                        modpath="input.func_w0401_package",
+                    ),
+                    FileItem(
+                        name="input.func_w0401_package.thing2",
+                        filepath=str(package_path / "thing2.py"),
+                        modpath="input.func_w0401_package",
+                    ),
+                ],
+            )
+
+        assert "cyclic-import" in linter.stats.by_msg
+
+    @pytest.mark.needs_two_cores
+    @patch("pylint.checkers.imports.ImportsChecker.close")
+    def test_cyclic_import_parallel_disabled_globally(self, mock: MagicMock) -> None:
+        tests_dir = Path("tests")
+        package_path = Path("input") / "func_w0401_package"
+        linter = PyLinter(reporter=Reporter())
+        linter.register_checker(ImportsChecker(linter))
+        linter.disable("cyclic-import")
+
+        with _test_cwd(tests_dir):
+            check_parallel(
+                linter,
+                jobs=2,
+                files=[
+                    FileItem(
+                        name="input.func_w0401_package.all_the_things",
+                        filepath=str(package_path / "all_the_things.py"),
+                        modpath="input.func_w0401_package",
+                    ),
+                    FileItem(
+                        name="input.func_w0401_package.thing2",
+                        filepath=str(package_path / "thing2.py"),
+                        modpath="input.func_w0401_package",
+                    ),
+                ],
+            )
+
+        mock.assert_not_called()
+        assert "cyclic-import" not in linter.stats.by_msg
+
+    @pytest.mark.needs_two_cores
+    def test_cyclic_import_parallel_disabled_locally(self) -> None:
+        tests_dir = Path("tests")
+        package_path = Path("input") / "func_noerror_cycle"
+        linter = PyLinter(reporter=Reporter())
+        linter.register_checker(ImportsChecker(linter))
+
+        with _test_cwd(tests_dir):
+            check_parallel(
+                linter,
+                jobs=2,
+                files=[
+                    FileItem(
+                        name="input.func_noerror_cycle.a",
+                        filepath=str(package_path / "a.py"),
+                        modpath="input.func_noerror_cycle",
+                    ),
+                    FileItem(
+                        name="input.func_noerror_cycle.b",
+                        filepath=str(package_path / "b.py"),
+                        modpath="input.func_noerror_cycle",
+                    ),
+                ],
+            )
+
+        assert "cyclic-import" not in linter.stats.by_msg
